/*
 * $QNXLicenseC:
 * Copyright 2008, QNX Software Systems. 
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"). You 
 * may not reproduce, modify or distribute this software except in 
 * compliance with the License. You may obtain a copy of the License 
 * at: http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied.
 *
 * This file may contain contributions from others, either as 
 * contributors under the License or as licensors under other terms.  
 * Please review this entire file for other proprietary rights or license 
 * notices, as well as the QNX Development Suite License Guide at 
 * http://licensing.qnx.com/license-guide/ for other information.
 * $
 */





#include "startup.h"
#include "lapic.h"

#define write_code(d, s, l)		memmove((void *)(d), (void *)(s), l)

extern struct callout_rtn	apstart;
extern uint8_t				apstart16_start[];
extern uint8_t				apstart16_end[];
extern uint8_t				apstart32[];
extern uint8_t				apflip[];
extern uint8_t				*apstart32_ptr;

static uint8_t		cpuid_remap[32];


#include <_pack1.h>

struct far_jmp {
	uint8_t			opcode;
	unsigned short	off;
	unsigned short	seg;
};


static void
smp_cmd(int id, unsigned long cmd) {
	unsigned long	val;
	int n;

	apic_write(LAPIC_ESR, 0);	// Clear any accumulated errors.

	// Set target APIC ID
	val = (((unsigned long)id) << 24) | (apic_read(LAPIC_ICRH) & 0x00ffffff);
	apic_write(LAPIC_ICRH, val);

	// Send command
	val = cmd | (apic_read(LAPIC_ICRL) & 0xfff33000);
	apic_write(LAPIC_ICRL, val);

	for(n = 10000 ; n ; --n) {
		val = apic_read(LAPIC_ICRL);
		if((val & 0x1000) == 0)
			break;
	}
}

static void
set_cpuid(unsigned id) {
	unsigned	idreg = apic_read(LAPIC_ID) & 0xf0ffffff;

	apic_write(LAPIC_ID, idreg | (id << 24));
}

	
static void
pause(int ms) {
	// Can't use timer hardware - CPU coming up reads/writes it as well
	// and gets confused. Maybe use the time stamp counter, if present.
	static volatile long cnt;

	while(ms--) {
		for(cnt = 10000 ; cnt ; --cnt) {
		}
	}
}

void
apic_add_cpuid(unsigned num_cpus, unsigned id) {
	unsigned	i;

	//sort the ID's in increasing order, to make it
	//easier to remap things later on.
	for(i = num_cpus; i > 0; --i) {
		if(cpuid_remap[i-1] < id) break;
		cpuid_remap[i] = cpuid_remap[i-1];
	}
	cpuid_remap[i] = id;
}


// Perform board specific initialization of the smp_entry section of
// the system page.
void
apic_smp_init(struct smp_entry *smp, unsigned num_cpus) {
	unsigned				size;
	paddr32_t				apstart;
	unsigned 				id;
	struct far_jmp			*jmp;

	// If the BSP is not CPU 0, make it so

	id = apic_id();
	if(id != 0) {

		// Check and see if somebody else is CPU 0.

		if(cpuid_remap[0] == 0) {
			extern volatile int		cpu_starting;
	
			cpu_starting = id;
			apstart32_ptr = (void *)apflip;
	
			size = (uintptr_t)apstart16_end - (uintptr_t)apstart16_start;
			/*
				Find a space in low memory where we can put the 16 bit startup
				code. We only need it while this function is executing, so
				we don't need to bother actually allocating it.
			*/
			apstart = find_ram(size, 1, 0, 0);
			if(apstart >= 0x100000) {
				crash("No low memory for AP startup\n");
			}
			write_code(apstart, apstart16_start, size);
		
			// We cream the first few real mode interrupt vectors to get a page aligned
			// addr to jump to.
			jmp = (void *)0;
			jmp->opcode = 0xea;	/* JMP FAR direct */
			jmp->off = apstart & 0x000f;
			jmp->seg = apstart >> 4;
	
			// Send Init command and wait at least 10 msec.
			smp_cmd(0, ICR_INIT);
			pause(10);
		
			// Send startup command at vector 00 (spec says do it twice).
			smp_cmd(0, ICR_STARTUP);
			pause(1);
			smp_cmd(0, ICR_STARTUP);
	
			// Wait for AP to acknowledge ID change
			do {
			} while(cpu_starting != 0);
		}

		// Now we can set the CPUID of the BSP to zero.
		set_cpuid(0);
	}
}


// Get the 'cpu' processor executing at the 'start' address.
int
apic_smp_start(unsigned cpu, void (*start)(void)) {
	unsigned				size;
	paddr32_t				apstart;
	unsigned 				id;
	struct far_jmp			*jmp;

	if(cpu == 1) {
		//
		// First time in
		//
		lsp.smp.p->cpu = (uintptr_t)lapic_vaddr;

		apstart32_ptr = (void *)start;
		/*
			Find a space in low memory where we can put the 16 bit startup
			code. We only need it while this function is executing, so
			we don't need to bother actually allocating it.
		*/
		size = (uintptr_t)apstart16_end - (uintptr_t)apstart16_start;
		apstart = find_ram(size, 1, 0, 0);
		if(apstart >= 0x100000) {
			crash("No low memory for AP startup\n");
		}
		write_code(apstart, apstart16_start, size);
	
		// We cream the first few real mode interrupt vectors to get a page aligned
		// addr to jump to.
		jmp = (void *)0;
		jmp->opcode = 0xea;	/* JMP FAR direct */
		jmp->off = apstart & 0x000f;
		jmp->seg = apstart >> 4;
	
		// Enable the local APIC
		apic_write(LAPIC_SVR, apic_read(LAPIC_SVR) | SVR_APIC_ENABLED);
	}

	id = cpuid_remap[cpu];

	// Send Init command and wait at least 10 msec.
	smp_cmd(id, ICR_INIT);
	pause(10);

	// Send startup command at vector 00 (spec says do it twice).
	smp_cmd(id, ICR_STARTUP);
	pause(1);
	smp_cmd(id, ICR_STARTUP);

	return 1;
}


// Perform any necessary adjustment of the cpu number
unsigned
apic_smp_adjust_num(unsigned cpu) {
	set_cpuid(cpu);
	return cpu;
}

__SRCVERSION("apic_smp.c $Rev: 1 $");
