/*
 * $QNXLicenseC:
 * Copyright 2007, 2008, QNX Software Systems. 
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"). You 
 * may not reproduce, modify or distribute this software except in 
 * compliance with the License. You may obtain a copy of the License 
 * at: http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied.
 *
 * This file may contain contributions from others, either as 
 * contributors under the License or as licensors under other terms.  
 * Please review this entire file for other proprietary rights or license 
 * notices, as well as the QNX Development Suite License Guide at 
 * http://licensing.qnx.com/license-guide/ for other information.
 * $
 */


/*
;
; This is a kernel hardware callout.
; It may be called from within an interrupt handler.
; This code MUST be position independant.
;
*/

#include "callout.ah"

/*
 * Interrupts are always off in this code.
 *
 * Note that interrupt_id_* and interrupt_eoi_* are not actually
 * called by the kernel. For performance reasons, they are instead
 * copied into place and intermixed with other kernel code. As such,
 * they don't follow normal calling conventions. They should just fall out 
 * their bottoms when done rather than attempting to do a return 
 * instruction. The 'id' routine should return the interrupt level in
 * the EAX register. 
 * By turning on various INTR_GENFLAG_* bits in the intrinfo_entry,
 * the following values can be loaded before entry to this code.
 * 
 *		syspage_ptr			=> EBX
 *		intrinfo_entry		=> ECX
 *		intr mask count		=> EDX (only available in EOI routine)
 *
 * If the INTR_FLAG_CPU_FAULT bit is on, the eoi routine should set
 * the EAX register to the following value to indicate the exception
 * that has occurred:
 * 		EAX = (fault_number << 16) | (signal_code << 8) | (signal_number)
 * Set EAX to zero if no exception has occured. If the exception is memory
 * related, set EDX to referenced address that caused the exception, zero
 * if there is no address.
 */

// syspage_paddr, vaddr, rtn_off, rw_off, data, rtn
// 4              8      0xc      0x10    0x14  0x18

patch_apic2:
    movl    0x4(%esp),%eax			// get paddr of routine
    addl    0xc(%esp),%eax			// ...
    movl    0x8(%esp),%edx			// get vaddr of R/W data
    addl    0x10(%esp),%edx			// ...
    movl    %edx,0x6(%eax)			// patch code with R/W vaddr
	// Fall through
patch_apic1:
    movl    0x4(%esp),%eax			// get paddr of routine
    addl    0xc(%esp),%eax			// ...
	movl	0x14(%esp),%edx			// get base info
	
    movl    %edx,0x1(%eax)			// patch code with APIC vaddr
    ret       

#define LVT_SPACING_SHIFT	4

mask_data:
	.long	4

/*
; 
; int interrupt_id()
;
; Returns: 0 - it's not a glitch   -1 - it's a glitch
;
*/
CALLOUT_START(interrupt_id_apic, 0, 0)
	//Nothing to do
CALLOUT_END(interrupt_id_apic)

/*
; 
; void interrupt_eoi
;
; Send an end-of-interrupt to the PIC.
;
*/
CALLOUT_START(interrupt_eoi_apic, 0, patch_apic1)
    movl	$0x12345678,%edx		// get APIC base vaddr (patched)
	movl	$0,LAPIC_EOI(%edx)		// Intel wants a zero written
CALLOUT_END(interrupt_eoi_apic)

/*
;
; int interrupt_unmask(volatile struct syspage_entry *syspage, int intr)
;
; Unmask the passed hardware interrupt number. 
;
*/
CALLOUT_START(interrupt_unmask_apic, mask_data, patch_apic2)
    movl    $0x12345678,%edx		// get APIC base vaddr (patched)
    movl    $0x12345678,%eax		// get R/W mask vaddr (patched)
	movl	0x8(%esp),%ecx
	orl		%ecx,%ecx
	jl		1f
	shl		$LVT_SPACING_SHIFT,%ecx
	andl	$~LVT_MASKED,LAPIC_LTR(%edx,%ecx,1)
	shr		$LVT_SPACING_SHIFT,%ecx
	movl	$~1,%edx
	shl		%cl,%edx
	andl	%edx,0(%eax)
	jmp		2f
1:	
	// sync with mask
	movl	0(%eax),%ecx
	movl	$6,%eax
3:
	shr		$1,%ecx
	jc		4f
	andl	$~LVT_MASKED,LAPIC_LTR(%edx)
	jmp		5f
4:	
	orl		$LVT_MASKED,LAPIC_LTR(%edx)
5:
	addl	$(1<<LVT_SPACING_SHIFT),%edx
	decl	%eax
	jne		3b

2:
    xorl    %eax,%eax
    ret       
CALLOUT_END(interrupt_unmask_apic)

/*
;
; int interrupt_mask(volatile struct syspage_entry *syspage, int intr)
;
; Mask the passed hardware interrupt number.
;
*/
CALLOUT_START(interrupt_mask_apic, mask_data, patch_apic2)
    movl    $0x12345678,%edx		// get APIC base vaddr (patched)
    movl    $0x12345678,%eax		// get R/W mask vaddr (patched)
	movl	0x8(%esp),%ecx
	orl		%ecx,%ecx
	jl		1f
	shl		$LVT_SPACING_SHIFT,%ecx
	orl		$LVT_MASKED,LAPIC_LTR(%edx,%ecx,1)
	shr		$LVT_SPACING_SHIFT,%ecx
	movl	$1,%edx
	shl		%cl,%edx
	orl		%edx,0(%eax)
	jmp		2f
1:	
	// sync with mask
	movl	0(%eax),%ecx
	movl	$6,%eax
3:
	shr		$1,%ecx
	jc		4f
	andl	$~LVT_MASKED,LAPIC_LTR(%edx)
	jmp		5f
4:	
	orl		$LVT_MASKED,LAPIC_LTR(%edx)
5:
	addl	$(1<<LVT_SPACING_SHIFT),%edx
	decl	%eax
	jne		3b

2:
    xorl    %eax,%eax
    ret       
CALLOUT_END(interrupt_mask_apic)

/*
;
; int interrupt_config(volatile struct syspage_entry *syspage,
                        struct intrinfo_entry *info, int intr)
;
; Return configuration information for interrupt
;
;
*/
CALLOUT_START(interrupt_config_apic, 0, 0)
    xor     %eax,%eax
    ret
CALLOUT_END(interrupt_config_apic)

/*
; 
; int interrupt_id()
;
; Returns: 0 - it's not a glitch   -1 - it's a glitch
;
*/
CALLOUT_START(interrupt_id_apic_spurious, 0, 0)
	mov		  $-1,%eax		
CALLOUT_END(interrupt_id_apic_spurious)

/*
; 
; void interrupt_eoi
;
; Send an end-of-interrupt to the PIC.
;
*/
CALLOUT_START(interrupt_eoi_apic_spurious, 0, 0)
CALLOUT_END(interrupt_eoi_apic_spurious)

/*
;
; int interrupt_unmask(volatile struct syspage_entry *syspage, int intr)
;
; Unmask the passed hardware interrupt number. 
;
*/
CALLOUT_START(interrupt_unmask_apic_spurious, 0, 0)
    xorl      %eax,%eax
    ret       
CALLOUT_END(interrupt_unmask_apic_spurious)

/*
;
; int interrupt_mask(volatile struct syspage_entry *syspage, int intr)
;
; Mask the passed hardware interrupt number.
;
*/
CALLOUT_START(interrupt_mask_apic_spurious, 0, 0)
	mov		  $22,%eax		// return EINVAL
    ret       
CALLOUT_END(interrupt_mask_apic_spurious)

/*
;
; int interrupt_config(volatile struct syspage_entry *syspage,
                        struct intrinfo_entry *info, int intr)
;
; Return configuration information for interrupt
;
*/
CALLOUT_START(interrupt_config_apic_spurious, 0, 0)
    mov     $INTR_CONFIG_FLAG_DISALLOWED,%eax
    ret
CALLOUT_END(interrupt_config_apic_spurious)
