/*
 * $QNXLicenseC:
 * Copyright 2007, 2008, QNX Software Systems. 
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"). You 
 * may not reproduce, modify or distribute this software except in 
 * compliance with the License. You may obtain a copy of the License 
 * at: http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied.
 *
 * This file may contain contributions from others, either as 
 * contributors under the License or as licensors under other terms.  
 * Please review this entire file for other proprietary rights or license 
 * notices, as well as the QNX Development Suite License Guide at 
 * http://licensing.qnx.com/license-guide/ for other information.
 * $
 */


#include "startup.h"

//
// Intel MultiProcessor Specification Routines
//

#define MPFPS_SIGNATURE	('_' | ('M'<<8) | ((long)'P'<<16) | ((long)'_'<<24))

#define MPFPS_ADDR1	((paddr32_t) (640L*1024-1024))
#define MPFPS_ADDR2	((paddr32_t) 16L* *(unsigned short *)0x40e)
#define MPFPS_ADDR3	((paddr32_t) 0xf0000)


#include <_pack1.h>

struct smp_fps {
	unsigned long	signature;
	paddr32_t		config;
	unsigned char	length;
	unsigned char	spec_rev;
	unsigned char	checksum;
	unsigned char	features[5];
};

struct smp_hdr {
	unsigned long	signature;
	unsigned short	base_length;
	unsigned char	spec_rev;
	unsigned char	checksum;
	char			oem_id[8];
	char			prod_id[12];
	paddr32_t		oem_table_ptr;
	unsigned short	oem_table_size;
	unsigned short	entry_count;
	paddr32_t		lapic_addr;
	unsigned short	extended_length;
	unsigned char	extended_checksum;
	char			reserved[1];
};


#define MP_PROCESSOR	0
#define MP_BUS			1
#define MP_IOAPIC		2
#define MP_INTSRC		3
#define MP_LINTSRC		4
struct smp_processor {
	unsigned char	type;			// Type 0
	unsigned char	id;
	unsigned char	ver;
	unsigned char	flags;
	unsigned long	signature;
	unsigned long	features;
	char			reserved[8];
};


struct smp_bus {
	unsigned char	type;			// Type 1
	unsigned char	id;
	char			bus_type[6];
};

struct smp_ioapic {
	unsigned char	type;			// Type 2
	unsigned char	id;
	unsigned char	ver;
	unsigned char	flags;
	unsigned long	addr;
};

struct smp_intsrc {
	unsigned char	type;			// Type 3
	unsigned char	int_type;
	unsigned short	flags;
	unsigned char	source_bus_id;
	unsigned char	source_bus_intr;
	unsigned char	dest_apic_id;
	unsigned char	dest_apic_intin;
};

struct smp_lintsrc {
	unsigned char	type;			// Type 4
	unsigned char	int_type;
	unsigned short	flags;
	unsigned char	source_bus_id;
	unsigned char	source_bus_intr;
	unsigned char	dest_apic_id;
	unsigned char	dest_apic_intin;
};

static const uint8_t entry_size[] = {
	sizeof(struct smp_processor),
	sizeof(struct smp_bus),
	sizeof(struct smp_ioapic),
	sizeof(struct smp_intsrc),
	sizeof(struct smp_lintsrc),
};


//
// Search for the "MP Floating Pointer Structure". It returns the offset
// of the structure. The structure is always on a 16 byte boundry and
// is 16 bytes long. It is located in the first megabyte of memory.
//
// Both start and length will be 16 byte aligned.
//
static struct smp_fps *
smp_scan_fps(paddr32_t beg, paddr32_t end) {
	static struct smp_fps buf;

	if(debug_flag > 4) {
		kprintf("Scan for MP from %l to %l\n", beg, end);
	}

	while(beg < end) {
		memmove(&buf, (void *)beg, sizeof(buf));
		if(buf.signature == MPFPS_SIGNATURE  &&
		  (buf.length == 1)  &&
		  (buf.spec_rev == 1  ||  buf.spec_rev == 4)  &&
		  (calc_cksum(&buf, 16) == 0)) {
			if(debug_flag > 4) {
				kprintf("Found MP Floating Structure Pointer at %l\n", beg);
			}
			return &buf;
		}
		beg += 16;
	}

	return NULL;
}


//
// Look in the 3 places the spec says the pointer structure can be.
// Note: We assume 640K of base memory. It is possible a system may
//       be compliant with only 512K. However, it is unlikely.
//
static struct smp_fps *
smp_get_fps(void) {
	struct smp_fps	*addr;

	addr = smp_scan_fps(MPFPS_ADDR1, MPFPS_ADDR1+1024);
	if(addr != NULL) return addr;
	addr = smp_scan_fps(MPFPS_ADDR2, MPFPS_ADDR2+1024);
	if(addr != NULL) return addr;
	addr = smp_scan_fps(MPFPS_ADDR3, MPFPS_ADDR3+0x10000);

	return addr;
}


// Return the number of CPU's on the system.
unsigned
imps_smp_num_cpu(void) {
	struct smp_fps				*fps;
	int							n;
	paddr32_t					addr;
	unsigned					num_cpus;
	union {
		uint8_t 				type;
		struct smp_hdr			hdr;
		struct smp_processor	processor;
		struct smp_bus			bus;
		struct smp_ioapic		ioapic;
		struct smp_intsrc		intsrc;
		struct smp_lintsrc		lintsrc;
	} 						cfg;

	fps = smp_get_fps();
	if(fps == NULL) {
		return 1;
	}

	if(fps->config == 0) {
		kprintf("MP configuration table is missing.\n");
		return 1;
	}

	memmove(&cfg.hdr, (void *)fps->config, sizeof(cfg.hdr));
	if(debug_flag > 2) {
		kprintf("LAPIC addr: %l, #entries=%d\n", cfg.hdr.lapic_addr, cfg.hdr.entry_count);
	}
	apic_set_paddr(cfg.hdr.lapic_addr);

	num_cpus = 0;
	addr = fps->config + sizeof(cfg.hdr);
	for(n = cfg.hdr.entry_count ; n ; --n) {
		memmove(&cfg, (void *)addr, sizeof(cfg));
		if(cfg.type == MP_PROCESSOR) {
			if(debug_flag > 2) {
				kprintf("LAPIC id: %w", cfg.processor.id);
				if(cfg.processor.flags & 0x02) {
					kprintf(" (BSP)");
				}
				kprintf("\n");
			}
			apic_add_cpuid(num_cpus, cfg.processor.id);
			++num_cpus;
		}
		if(cfg.type >= sizeof(entry_size)/sizeof(entry_size[0])) {
			kprintf("Unknown SMP entry type %d.\n", cfg.type);
			break;
		}
		addr += entry_size[cfg.type];
	}
	
	/*
	*  Temp fix for Dell P4's. For some reason they decide about
	*  half way thru the startup that there isn't actually a processor.
	*  This should be revisted when we have time.
	*/
	if(num_cpus == 0) num_cpus = 1;

	return num_cpus;
}

__SRCVERSION("imps_smp.c $Rev: 1 $");
