/*
 * $QNXLicenseC:
 * Copyright 2008, QNX Software Systems. 
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"). You 
 * may not reproduce, modify or distribute this software except in 
 * compliance with the License. You may obtain a copy of the License 
 * at: http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied.
 *
 * This file may contain contributions from others, either as 
 * contributors under the License or as licensors under other terms.  
 * Please review this entire file for other proprietary rights or license 
 * notices, as well as the QNX Development Suite License Guide at 
 * http://licensing.qnx.com/license-guide/ for other information.
 * $
 */





#include "startup.h"

paddr32_t		gdt_paddr;

#define KERN_CODE_MAX	0xFFFFFFFF
#define KERN_DATA_MAX	0xFFFFFFFF

//
// Initialize gdt structure in the system page.
// This code is hardware independant and should not have to be
// changed by end users.
//
void init_gdt(unsigned idt_size) {
	struct x86_seg_descriptor_entry	*gdt = TOPTR(gdt_paddr);
	int 							i;
	uintptr_t						idtptr;
	
	//
	// Entry 1 points to the gdt table itself.
	//       2 points to the idt table.
	//
	set_desc32(gdt +  1, X86_TYPE_DATA_SEG|X86_DPL1, 0, (uintptr_t)lsp.syspage.p->un.x86.gdt, GDT_SIZE);
	//
	// Use the read-only copy for the IDT to avoid the Pentium F00F bug... 
	//
	idtptr = (uintptr_t)lsp.system_private.p->user_syspageptr + PTR_DIFF(lsp.syspage.p->un.x86.idt, lsp.system_private.p->kern_syspageptr);
	set_desc32(gdt +  2, X86_TYPE_DATA_SEG|X86_DPL1, 0, idtptr, idt_size);
	//
	// Entries 4..11 point to the cpupage entry for each processor
	//
	for(i = 0 ; i < lsp.syspage.p->num_cpu; ++i) {
		set_desc32(gdt + (i+4), X86_TYPE_DATA_SEG|X86_DPL3, 0, i*sizeof(struct cpupage_entry) + (uintptr_t)lsp.system_private.p->user_cpupageptr, sizeof(struct cpupage_entry));
	}

	//
	// Entry 15 points to the 1st page of physical memory which is only needed
	// for virtual 8086 mode.
	//
	set_desc32(gdt + 15, X86_TYPE_CODE_SEG|X86_DPL0, X86_G_BIT,
			(shdr->flags1 & STARTUP_HDR_FLAGS1_VIRTUAL) ? REAL_VADDR : 0, 4096);

	//
	// Entry 16 and 17 map the code and data for the kernel which runs at
	// privity 0.
	//
	set_desc32(gdt + 16, X86_TYPE_CODE_SEG|X86_DPL0, X86_D_BIT|X86_G_BIT, 0, KERN_CODE_MAX+1);
	set_desc32(gdt + 17, X86_TYPE_DATA_SEG|X86_DPL0, X86_B_BIT|X86_G_BIT, 0, KERN_DATA_MAX+1);
	
	//
	// Entry 18 and 19 map the code and data for ProcNto as a process
	// which runs at provity 1.
	set_desc32(gdt + 18, X86_TYPE_CODE_SEG|X86_DPL0|X86_CONFORMING_CODE, X86_D_BIT|X86_G_BIT, 0, KERN_CODE_MAX+1);
	set_desc32(gdt + 19, X86_TYPE_DATA_SEG|X86_DPL1, X86_B_BIT|X86_G_BIT, 0, KERN_DATA_MAX+1);

	//
	// Entry 20 is an LDT which points right into the GDT at entry 21.
	// User mode programs run segments in the LDT which is equivalent to
	// running with segments in the GDT of 21,22,23,24 and 25.
	//
	set_desc32(gdt + 20, X86_LDT_DESCRIPTOR|X86_PRESENT|X86_DPL0, 0, (uintptr_t)lsp.syspage.p->un.x86.gdt + 21*sizeof(*gdt), sizeof(*gdt) * 5);
	
	//
	// The user mode segments for code and data.
	//
	set_desc32(gdt + 24, X86_TYPE_CODE_SEG|X86_DPL0|X86_CONFORMING_CODE, X86_D_BIT|X86_G_BIT, 0, KERN_CODE_MAX+1);
	set_desc32(gdt + 25, X86_TYPE_DATA_SEG|X86_DPL3, X86_B_BIT|X86_G_BIT, 0, KERN_DATA_MAX+1);

	//
	// These are the set of 4 selectors that are used with SYSENTER/SYSEXIT
	// The first two are the kernel CS and SS, the last 2 are the user-mode
	// CS and SS.
	//
	set_desc32(gdt + 28, X86_TYPE_CODE_SEG|X86_DPL0, X86_D_BIT|X86_G_BIT, 0, KERN_CODE_MAX+1);
	set_desc32(gdt + 29, X86_TYPE_DATA_SEG|X86_DPL0, X86_B_BIT|X86_G_BIT, 0, KERN_DATA_MAX+1);
	set_desc32(gdt + 30, X86_TYPE_CODE_SEG|X86_DPL0|X86_CONFORMING_CODE, X86_D_BIT|X86_G_BIT, 0, KERN_CODE_MAX+1);
	set_desc32(gdt + 31, X86_TYPE_DATA_SEG|X86_DPL3, X86_B_BIT|X86_G_BIT, 0, KERN_DATA_MAX+1);
	//
	// Entries 32..39 point to the tss entry for each processor. They are setup
	// by the kernel
	//
}

__SRCVERSION("init_gdt.c $Rev: 1 $");
