/*
 * $QNXLicenseC:
 * Copyright 2007, 2008, QNX Software Systems. 
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"). You 
 * may not reproduce, modify or distribute this software except in 
 * compliance with the License. You may obtain a copy of the License 
 * at: http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied.
 *
 * This file may contain contributions from others, either as 
 * contributors under the License or as licensors under other terms.  
 * Please review this entire file for other proprietary rights or license 
 * notices, as well as the QNX Development Suite License Guide at 
 * http://licensing.qnx.com/license-guide/ for other information.
 * $
 */


#include "startup.h"
#include "multiboot.h"

//
// Initialize raminfo structure in the system page.
// This code is system dependant and may have to be
// changed by end users.
//
unsigned mem_scan;		// Global set via a command line option in main()

static void
my_add_ram(paddr64_t big_start, paddr64_t size) {
	// Find out how much conventional memory we've got.
	unsigned	conventional = (*(uint16_t *)0x413) * 1024;
	paddr_t		end;
	paddr_t		start;

#if _PADDR_BITS == 64
	start = big_start;
	end = start + size - 1;
#else
	// If we're compiled with a 32 bit paddr_t, restrict ourselves
	// to the first 4G.
	if(big_start > 0xffffffff) return;
	start = big_start;
	if((start + size) > 0xffffffff) {
		end = 0xffffffff;
	} else {
		end = start + size - 1;
	}
#endif

	// Deal with buggy bioses that tell us we've got memory
	// that actually belongs to the extended BIOS data area
	if((end < KILO(640)) && (conventional < (end + 1))) {
		end = conventional - 1;
	}

	// Deal with buggy bioses that tell us we've got memory
	// in the 640K -> 1M region
	if(!((start >= MEG(1)) || (end < KILO(640)))) {
		if(end < MEG(1)) end = KILO(640) - 1;
		if(start >= KILO(640)) start = MEG(1);
	}

	if(end >= start) {
		add_ram(start, end - start + 1);
	}

	// If we find memory above 64M then BIOS must be new so never scan
	if(mem_scan && (end > mem_scan)) {
		mem_scan = 0;
	}
}

//
// Memory located by QNX IPL.
//
static int 
ipl_ram() {
	int										found;
	const struct startup_info_mem_extended	*mem;
	paddr64_t								addr;
	paddr64_t								size;

	found = 0;
	mem = NULL;
	for (;;) {
		mem = find_startup_info(mem, STARTUP_INFO_MEM);
		if(mem == NULL) break;
		addr = mem->mem.addr;
		size = mem->mem.size;
		if(mem->mem.hdr.size >= sizeof(struct startup_info_mem_extended)) {
			addr |= (paddr64_t)mem->addr_hi << 32;
			size |= (paddr64_t)mem->size_hi << 32;
		}
		my_add_ram(addr, size);
		found = 1;
	} 

	return found;
} 

//
// Memory located by multiboot loader
//
static int
multiboot_ram() {
	struct mb_info	*info;

	if(boot_regs->eax == MB_SIGNATURE) {
		info = (void *)boot_regs->ebx;

		if(info->flags & MB_INFO_FLAG_MMAP_VALID) {
			struct mb_info_mmap	*mm_info	= (void *)info->mmap_addr;
			unsigned len;

			len = info->mmap_length;
			if(len > 0) {
				do {
					unsigned		size;

					if(mm_info->type == MB_INFO_MMAP_TYPE_RAM) {
						my_add_ram(
							mm_info->base_addr_low | (paddr64_t)mm_info->base_addr_high << 32,
							mm_info->length_low | (paddr64_t)mm_info->length_high << 32
						);
					}
					size = mm_info->size + 4;
					mm_info = (void *)((uintptr_t)mm_info + size);
					len -= size;
				} while(len != 0);

				return 1;
			}
		}

		if(info->flags & MB_INFO_FLAG_MEM_VALID) {
			my_add_ram(0, info->mem_lower * 1024);
			my_add_ram(MEG(1), info->mem_upper * 1024);

			return 1;
		}
	}
	return 0;
}

void
init_raminfo() {
	// Note that add_ram() removes overlapping memory additions.
	if(ipl_ram() || multiboot_ram()) {
		if(mem_scan) {
			x86_scanmem(MEG(16), MEG(256)-1);
		}
	} else {
		x86_scanmem(0, MEG(16)-1);
	}
}

__SRCVERSION("init_raminfo.c $Rev: 1 $");
